%{
#include <ctype.h>
#include <stdio.h>
#define MAX_ID_LEN 100
#define MAX_TABLE_LEN 100
typedef struct {char name[MAX_ID_LEN]; int value; int isDeclared; int isInitialized;}entry;	/*symbol table entry type*/
int declarations=0;			/*number of valid declarations*/
int lines=0;				/*number of lines processed*/
int n=0;					/*actual length of symbol table*/
int temp_n=0;				/*previous length of symbol table (temporary)*/
int i=0;
entry table[MAX_TABLE_LEN];
%}

%union{
		int const_type;
		entry *id_type;
		char op_type;
		}

%token VAR
%token <const_type> CONST
%token <id_type> ID
%token <id_type> NEW_ID
%type <const_type> term
%type <const_type> expr
%type <const_type> asm_rhs
%type <op_type> operator

%left '+' '-' '*'
%right '='

%%
prog		:	decl_list asm_list
		;

decl_list	:	decl ';' decl_list | /*e*/
		;

decl		:	VAR NEW_ID	{	$2->isDeclared=1;declarations++;	}
				| VAR ID	{	yyerror("Error:ID already defined:");
								printf("--At Line %d ID name: %s\n",lines+1,$2->name);
								yyerrok;	}
				| VAR error	{	yyerror("Error:Invalid ID.");
								printf("--At Line %d\n",lines+1);
								yyerrok;	}
		;

asm_list	:	asm ';' asm_list | /*e*/
				| error	';'	asm_list	{	yyerror("Invalid assignment.");
											printf("--At Line %d\n",lines+1);
											yyerrok;	}
		;

asm			:	ID '=' asm_rhs			{	$1->value=$3;	
											$1->isInitialized=1;	}
				| NEW_ID '=' asm_rhs	{	yyerror("Error:ID not declared:");
											printf("--At Line: %d ID name: %s\n",lines+1,$1->name);
											n--;
											yyerrok;	}
		;
		
asm_rhs		:	expr	{$$=$1;}
				| term	{$$=$1;}
		;

expr		:	ID operator term	{	if($1->isInitialized==0)
											{	yyerror("Warning: ID value not initialized:");
												printf("--At Line: %d ID name: %s Action: Use default value 0\n",lines+1,$1->name);
												yyerrok;
											}
										switch($2)
											{	case '+': $$=$1->value+$3;break;
												case '-': $$=$1->value-$3;break;
												case '*': $$=$1->value*$3;break;
												case '=': yyerror("Error: operator can't be =:");
														printf("--At Line: %d Recovery: Evaluate as %d\n",lines+1,$3);
														$1->value=$3;
														$$=$3;
														yyerrok;
														break;
											}
									}
				| CONST operator term	{	yyerror("Error: CONST can't be LHS of expression:");
											printf("--At Line: %d Recovery: Use CONST %d as LHS operand\n",lines+1,$1);
											yyerrok;
											switch($2)
											{	case '+': $$=$1+$3;break;
												case '-': $$=$1-$3;break;
												case '*': $$=$1*$3;break;
												case '=': yyerror("Error: invalid expression:");
														printf("--At Line: %d Recovery: Evaluate as %d\n",lines+1,$3);
														$$=$3;
														yyerrok;
														break;
											}
										}
				| NEW_ID operator term	{	yyerror("Error:ID not declared:");
											printf("--At Line: %d ID name: %s\n",lines+1,$1->name);
											n--;
											yyerrok;	}
		;

operator	:	'+'		{$$='+';}
				| '-'	{$$='-';}
				| '*'	{$$='*';}
				| '='	{$$='=';}
		;
		
term		:	ID		{	if($1->isInitialized==0)
								{	yyerror("Warning: ID value not initialized:");
									printf("--At Line: %d ID name: %s Action: Use default value 0\n",lines+1, $1->name);
									yyerrok;
								}
							$$=$1->value;	}
				| CONST	{$$=$1;}
				| NEW_ID{	yyerror("Error:ID not declared:");
							printf("--At Line: %d ID name: %s\n",lines+1,$1->name);
							n--;
							yyerrok;	} 
		;
%%
#include "lex.yy.c"

yyerror(s)
char *s;
{
  printf ("%s\n\t\t", s);
}

main( argc, argv )
int argc;
char **argv;
    {
    ++argv, --argc;  /* skip over program name */
    if ( argc > 0 )
            yyin = fopen( argv[0], "r" );

    else
            yyin = stdin;
	yyparse();

	for (i=1;i<=n;i++)
	{
		if(table[i].isDeclared==1)
			printf("%s = %d \n",table[i].name, table[i].value);
	}
	printf("Total number of lines processed: %d\n", lines);
	printf("Total number of declarations: %d\n", declarations);
}