%{
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define MAX_ID_LEN 100
#define MAX_TABLE_LEN 100
#define MAX_PARAM_NUMBER 100
#define MAX_VAR_NUMBER 100
#define MAX_STMT_LEN 1000
#define MAX_STMT_NUMBER 100
#define MAX_NODE_NUMBER 1000
typedef struct							/*The temporary symbol table structure*/
 {
	char name[MAX_ID_LEN];
	int isDeclared;
	int isInitiated;
 }entry;

typedef struct							/*The main symbol table structure*/
 {
	char name[MAX_ID_LEN]; 
	int isDeclared;
	int paramNum; 
	struct
	{
		char paramName[MAX_ID_LEN];
		int paramVal;
	}param[MAX_PARAM_NUMBER];
	int varNum;
	struct
	{
		char varName[MAX_ID_LEN];
		int varVal;
	}var[MAX_VAR_NUMBER];
	int statemtNum;
	char statemt[MAX_STMT_NUMBER][MAX_STMT_LEN];
	char output[MAX_ID_LEN];
 }funcentry;

typedef struct							/*The node structure*/
 {
	int funcVal;						/*Synthesized, to count actual parameters, used by actual_para_list*/
	char stmtVal[MAX_STMT_LEN];			/*Synthesized, Content of statement, used by stmt and its sub nodes*/
 }node;


int clearTable();
int checkID(int, char *);
int checkFunc(char *, int);
int addFunc(char *);
int addParam(int, char *);
int addVar(int, char *);
int addStmt(int, char *);
node *initNode();

char mainFunction[MAX_ID_LEN]="undefined";			/*name of main function declared at the beginning*/
int finalMainFunction=-1;							/*index of final main function to be converted into C*/
int tableLen=0;										/*actual length of temp symbol table*/
int functableLen=0;									/*actual length of main symbol table*/
int currentfunc=0;									/*current function being parsed*/
int temp_n=0;
char tempstring[MAX_ID_LEN];
node tempnode[MAX_NODE_NUMBER];
int nodeNum=0;
int i=0;
int j=0;
FILE *fo;
entry table[MAX_TABLE_LEN];				/*The temporary symbol table used by LEX*/
funcentry functable[MAX_TABLE_LEN];		/*The main symbol table*/
%}


%union{
		int num_type;
		entry *id_type;
		node *node_type;
		}

%token MYBEGIN
%token END
%token INPUT
%token OUTPUT
%token WHILE
%token DO
%token FUNCTION
%token MAIN
%token VAR
%token GREATER_OR_EQUAL
%token LESS_OR_EQUAL
%token GREATER
%token LESS
%token NOT_EQUAL
%token <num_type> NUM
%token <id_type> ID
%type <node_type> function
%type <node_type> func_body
%type <node_type> para
%type <node_type> para_list
%type <node_type> var_list
%type <node_type> stmt_list
%type <node_type> stmt
%type <node_type> asmt
%type <node_type> whil
%type <node_type> expr
%type <node_type> func_call
%type <node_type> actual_para
%type <node_type> actual_para_list
%type <node_type> relat
%type <node_type> whilstmt_list

%left '+' '-' '*'
%right '='

%%
prog		:
			MAIN FUNCTION '=' ID ';'
			{	
				strcpy(mainFunction, $4->name);
				clearTable();
			}
			func_list
		;

func_list	:
			function
			{
				clearTable();
			}
			func_list 
		|	/*e*/
		;

function	:
			FUNCTION ID 
			{
				$$=initNode();
				if(addFunc($2->name)==1)
					{
						currentfunc=functableLen-1;
					}
				else
					yyerror("Cannot declare function, function name already used.");
				yyerrok;
			}
			'('	INPUT para ')' MYBEGIN func_body OUTPUT ID ';'
			{
				if(checkID(currentfunc, $11->name)>0)
					strcpy(functable[currentfunc].output, $11->name);
				else
				{
					yyerror("Output not declared");
					strcpy(functable[currentfunc].output, "0/*output not defined, use 0*/");
				}
				yyerrok;
			}
			END
		;

para		:
			para_list
		|	/*e*/
			{
				functable[currentfunc].paramNum=0;
			}
		;

para_list	:
			ID ',' para_list
			{
				if(addParam(currentfunc, $1->name)==0)
					yyerror("Max number of parameter reached");
				yyerrok;
			}
		|	ID
			{
				if(addParam(currentfunc, $1->name)==0)
					yyerror("Max number of parameter reached");
				yyerrok;
			}
		;

func_body	:
			VAR	var_list ';' stmt_list
			{}
		;

var_list	:
			ID ',' var_list
			{
				if(checkID(currentfunc, $1->name)!=1)
				{
					if(addVar(currentfunc, $1->name)==0)
						yyerror("Max number of var reached");
					yyerrok;
				}
				else
				{
					yyerror("Var already defined in parameter");
					yyerrok;
				}
			}
		|	ID
			{
				if(checkID(currentfunc, $1->name)!=1)
				{
					if(addVar(currentfunc, $1->name)==0)
						yyerror("Max number of var reached");
					yyerrok;
				}
				else
				{
					yyerror("Var already defined in parameter");
					yyerrok;
				}
			}
		;

stmt_list	:
			stmt ';' stmt_list
			{
				if(addStmt(currentfunc, $1->stmtVal)==0)
					yyerror("Max number of statement reached");
				yyerrok;
			}
		|	stmt ';'
			{
				if(addStmt(currentfunc, $1->stmtVal)==0)
					yyerror("Max number of statement reached");
				yyerrok;
			}
		;

stmt		:
			asmt
			{
				$$=initNode();
				strcpy($$->stmtVal,$1->stmtVal);
			}
		|	whil
			{
				$$=initNode();
				strcpy($$->stmtVal,$1->stmtVal);
			}
		;

asmt		:
			ID '=' expr
			{
				$$=initNode();
				if(checkID(currentfunc, $1->name)==0)
					{
						strcpy($$->stmtVal,"/*Undeclared ID used in assignment, declaration added by parser*/\n");
						strcpy($$->stmtVal, strcat($$->stmtVal, $1->name));
						addVar(currentfunc, $1->name);
						yyerror("Undeclared ID used in assignment"); yyerrok;
					}
				else
					strcpy($$->stmtVal,$1->name);
				strcpy($$->stmtVal, strcat($$->stmtVal, " = "));
				strcpy($$->stmtVal, strcat($$->stmtVal, $3->stmtVal));
			}
		|	expr
			{
				$$=initNode();
				strcpy($$->stmtVal,"/*Isolated expression: must be evaluated in a statement*/\n");
				strcpy($$->stmtVal, strcat($$->stmtVal, "/*"));
				strcpy($$->stmtVal, strcat($$->stmtVal, $1->stmtVal));
				strcpy($$->stmtVal, strcat($$->stmtVal, "*/"));
				yyerror("Isolated expression");
				yyerrok;
			}
		|	ID '=' '=' expr
			{
				$$=initNode();
				strcpy($$->stmtVal,"/*Invalid assignment using ==, corrected by parser*/\n");
				strcpy($$->stmtVal, strcat($$->stmtVal, $1->name));
				strcpy($$->stmtVal, strcat($$->stmtVal, "="));
				strcpy($$->stmtVal, strcat($$->stmtVal, $4->stmtVal));
				yyerror("Invalid assignment");
				yyerrok;
			}
		;

expr		:
			expr '+' expr
			{
				$$=initNode();
				strcpy($$->stmtVal,strcat($1->stmtVal, " + "));
				strcpy($$->stmtVal,strcat($$->stmtVal, $3->stmtVal));	
			}
		|	expr '-' expr
			{
				$$=initNode();
				strcpy($$->stmtVal,strcat($1->stmtVal, " - "));
				strcpy($$->stmtVal,strcat($$->stmtVal, $3->stmtVal));	
			}
		|	'(' expr ')'
			{
				$$=initNode();
				strcpy($$->stmtVal,strcat("(", $2->stmtVal));
				strcpy($$->stmtVal,strcat($$->stmtVal, ")"));
			}
		|	expr '*' expr
			{
				$$=initNode();
				strcpy($$->stmtVal,strcat($1->stmtVal, " * "));
				strcpy($$->stmtVal,strcat($$->stmtVal, $3->stmtVal));	
			}
		|	ID
			{
				$$=initNode();
				if(checkID(currentfunc, $1->name)==0)
				{
					strcpy($$->stmtVal,"/*Undeclared ID used in assignment, declared by parser*/");
					strcpy($$->stmtVal, strcat($$->stmtVal, $1->name));
					addVar(currentfunc, $1->name);
					yyerror("Undeclared ID used in expression"); yyerrok;
				}
				else
				{
					strcpy($$->stmtVal,$1->name);
				}
			}
		|	func_call
			{
				$$=initNode();
				strcpy($$->stmtVal,$1->stmtVal);
			}
		|	NUM
			{
				$$=initNode();
				strcpy($$->stmtVal,itoa($1, tempstring, 10));
			}
		;

func_call	:
			ID '(' actual_para ')'
			{
				$$=initNode();
				if(checkFunc($1->name, $3->funcVal)>=0)
					{
						strcpy($$->stmtVal,strcat($1->name, "("));
						strcpy($$->stmtVal,strcat($$->stmtVal, $3->stmtVal));
						strcpy($$->stmtVal,strcat($$->stmtVal, ")"));
					}
				else
					{
						strcpy($$->stmtVal, "0/*Undeclared function in function call*/");
						yyerror("Undeclared function in function call");
						yyerrok;
					}
			}
		;

actual_para			:
			actual_para_list
			{
				$$=initNode();
				$$->funcVal=$1->funcVal;
				strcpy($$->stmtVal,$1->stmtVal);
			}
		|	/*e*/
			{
				$$=initNode();
				$$->funcVal=0;
				strcpy($$->stmtVal, "");
			}
		;

actual_para_list	:
			NUM ',' actual_para_list
			{
				$$=initNode();
				$$->funcVal=$3->funcVal+1;
				strcpy($$->stmtVal,strcat(itoa($1, tempstring, 10),", "));
				strcpy($$->stmtVal,strcat($$->stmtVal, $3->stmtVal));
			}
		|	NUM
			{
				$$=initNode();
				$$->funcVal=1;
				strcpy($$->stmtVal, itoa($1, tempstring, 10));
			}
		|	ID ',' actual_para_list
			{
				$$=initNode();
				if(checkID(currentfunc, $1->name)>0)
				{
					strcpy($$->stmtVal,$1->name);
				}
				else
				{
					strcpy($$->stmtVal,"/*Undeclared ID used in assignment, declared by parser*/");
					strcpy($$->stmtVal, strcat($$->stmtVal, $1->name));
					addVar(currentfunc, $1->name);
					yyerror("Undeclared ID used in function call"); yyerrok;
				}
				strcpy($$->stmtVal,strcat($$->stmtVal, ", "));
				strcpy($$->stmtVal,strcat($$->stmtVal, $3->stmtVal));
				$$->funcVal=$3->funcVal+1;
			}
		|	ID
			{
				$$=initNode();
				if(checkID(currentfunc, $1->name)>0)
				{
					strcpy($$->stmtVal,$1->name);
				}
				else
				{
					strcpy($$->stmtVal,"/*Undeclared ID used in assignment, declared by parser*/");
					strcpy($$->stmtVal, strcat($$->stmtVal, $1->name));
					addVar(currentfunc, $1->name);
					yyerror("Undeclared ID used in function call"); yyerrok;
				}
				$$->funcVal=1;
			}
		;

whil		:
			WHILE '(' expr relat expr ')' DO MYBEGIN whilstmt_list END
			{
				$$=initNode();
				strcpy($$->stmtVal,"while (");
				strcpy($$->stmtVal, strcat($$->stmtVal,$3->stmtVal));
				strcpy($$->stmtVal, strcat($$->stmtVal, $4->stmtVal));
				strcpy($$->stmtVal, strcat($$->stmtVal, $5->stmtVal));
				strcpy($$->stmtVal, strcat($$->stmtVal, ")\n{\n"));
				strcpy($$->stmtVal, strcat($$->stmtVal, $9->stmtVal));
				strcpy($$->stmtVal, strcat($$->stmtVal, "}\n"));
			}
		;

relat		:
			GREATER_OR_EQUAL	{$$=initNode();strcpy($$->stmtVal, " >= ");}
		|	LESS_OR_EQUAL		{$$=initNode();strcpy($$->stmtVal, " <= ");}
		|	NOT_EQUAL			{$$=initNode();strcpy($$->stmtVal, " != ");}
		|	GREATER				{$$=initNode();strcpy($$->stmtVal, " > ");}
		|	LESS				{$$=initNode();strcpy($$->stmtVal, " < ");}
		|	'='					{$$=initNode();strcpy($$->stmtVal, " == ");}
		;

whilstmt_list	:
			stmt ';' whilstmt_list
			{
				$$=initNode();
				strcpy($$->stmtVal, strcat($1->stmtVal, ";\n"));
				strcpy($$->stmtVal, strcat($1->stmtVal, $3->stmtVal));

			}
		|	stmt ';'
			{
				$$=initNode();
				strcpy($$->stmtVal, $1->stmtVal);
				strcpy($$->stmtVal, strcat($1->stmtVal, ";\n"));
			}
		;

%%
#include "lex.yy.c"

int clearTable()
{
	for(i=0;i<MAX_TABLE_LEN;i++)
		{
			strcpy(table[i].name, "");
		}
	tableLen=0;
	return 1;
}

int checkID(int funcindex, char *id)
{
	for(i=0;i<functable[funcindex].paramNum;i++)
		if(strcmp(functable[funcindex].param[i].paramName, id)==0)
			return 1;
	for(i=0;i<functable[funcindex].varNum;i++)
		if(strcmp(functable[funcindex].var[i].varName, id)==0)
			return 2;
	return 0;
}

int checkFunc(char *id, int pNum)
{
	for(i=0;i<functableLen;i++)
		if((strcmp(functable[i].name, id)==0)&&(functable[i].paramNum==pNum))
			return i;
	return -1;
}

int addFunc(char *newfunc)
{
	for(i=0;i<functableLen;i++)
		if(strcmp(functable[i].name,newfunc)==0)
			return 0;
	strcpy(functable[functableLen].name,newfunc);
	functable[functableLen].isDeclared=1;
	clearTable();
	functableLen++;
	return 1;
}

int addParam(int funcindex, char *id)
{
	int n=functable[funcindex].paramNum;
	if(n<MAX_PARAM_NUMBER)
		{
			strcpy(functable[funcindex].param[n].paramName,id);
			(functable[funcindex].paramNum)++;
			return 1;
		}
	else return 0;
}

int addVar(int funcindex, char *id)
{
	int n=functable[funcindex].varNum;
	if(n<MAX_VAR_NUMBER)
		{
			strcpy(functable[funcindex].var[n].varName,id);
			(functable[funcindex].varNum)++;
			return 1;
		}
	else return 0;
}

int addStmt(int funcindex, char *stmtContent)
{
	int n=functable[funcindex].statemtNum;
	if(n<MAX_STMT_NUMBER)
		{
			strcpy(functable[funcindex].statemt[n], stmtContent);
			(functable[funcindex].statemtNum)++;
			return 1;
		}
	else return 0;
}

node *initNode()
{
	nodeNum++;
	return &tempnode[nodeNum];
}

yyerror(s)
char *s;
{
  printf ("/*%s\n\t\t*/", s);
}

main(argc, argv )
int argc;
char **argv;
{
    ++argv, --argc;  /* skip over program name */
	if ( argc > 0 )
            yyin = fopen( argv[0], "r" );
    else
            yyin = stdin;
	yyparse();
	
	fo=fopen("output.c", "w");
	fprintf(fo, "/*Output C file translated by YACC\n");
	fprintf(fo, " *CSE 244 Final Project\n");
	fprintf(fo, " *Haibei Zhang Dec 17, 2002\n");
	fprintf(fo, " */\n\n");
	fprintf(fo, "#include<stdlib.h>\n");
	fprintf(fo, "#include<stdio.h>\n");
	
	/*Print Functions*/
	for(i=0;i<functableLen;i++)
	{
		/*Print function declaration*/
		fprintf(fo, "int %s(", functable[i].name);
		if(functable[i].paramNum>0)
		{
			for(j=functable[i].paramNum-1;j>0;j--)
				fprintf(fo, "int %s, ", functable[i].param[j].paramName);
			fprintf(fo, "int %s", functable[i].param[0].paramName);
		}
		fprintf(fo, ")\n");

		/*Print leading brachet*/
		fprintf(fo, "{\n");
		
		/*Print declaration*/
		fprintf(fo, "int ");
		if(functable[i].varNum>0)
		{
			for(j=functable[i].varNum-1;j>0;j--)
				fprintf(fo, "%s, ", functable[i].var[j].varName);
			fprintf(fo, "%s", functable[i].var[0].varName);
			fprintf(fo, ";\n\n");
		}

		/*Print Statements*/
		if(functable[i].statemtNum>0)
		{
			for(j=functable[i].statemtNum-1;j>0;j--)
				fprintf(fo, "%s; \n", functable[i].statemt[j]);
			fprintf(fo, "%s; \n", functable[i].statemt[0]);
		}

		/*Print return value*/
		fprintf(fo, "return %s;\n", functable[i].output);

		/*Print ending bracket*/
		fprintf(fo, "}\n\n\n");

		if(strcmp(functable[i].name, mainFunction)==0)
			finalMainFunction=i;
	}

	/*Print Main Function*/
	fprintf(fo, "main(argc, argv )\n");
	fprintf(fo, "int argc; char **argv;\n");
	fprintf(fo, "{\n");
	fprintf(fo, "++argv, --argc;  /* skip over program name */\n");
	if(finalMainFunction>=0)
		{
			fprintf(fo, "if ( argc == %d )  /* match argument count with main function */\n", functable[finalMainFunction].paramNum);
			fprintf(fo, "\tprintf(\"%c%c\\n\", %s(atoi(", '%','d', functable[finalMainFunction].name);
			if(functable[finalMainFunction].paramNum>0)
			{
				for(j=0;j<functable[finalMainFunction].paramNum-1;j++)
					fprintf(fo, "argv[%d], ",j);
				fprintf(fo, "argv[%d]",functable[finalMainFunction].paramNum-1);
			}
			fprintf(fo, ")));\n");
			fprintf(fo, "else\n");
			fprintf(fo, "\tprintf(\"Input Error: argument count does not match to main function %s\\n\");\n", functable[finalMainFunction].name);
		}
	else
		fprintf(fo, "/*Error: main function %s not declared*/\n", mainFunction);
	fprintf(fo, "}");	

}